import java.io.*;
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

import java.util.regex.*;

/**
 *	This class is an xml parsing handler class which extends the DefaultHandler
 *  in the org.xml.sax.helpers package.  It uses utility classes PathList, 
 *  ElementList and AttributeList to keep track of xml paths, elements and
 *  attributes respectively.  This class reads in an xml file specified as the 
 *  first command line argument and prints the results to standard out.  
 *  The methods are listed in alphabetical order except for the main method 
 *  which is listed first.
 *
 *  Usage: java s02q2 xmlFile 
 *
 *  @author Andrew McDowell
 */

public class s02q2 extends DefaultHandler
{	
	// Document locator for keeping track of start and end indexs
	private Locator locator;
	
	// PathList object used for keeping track of paths
	private static PathList paths = new PathList();
	
	// ElementList object used for keeping track of elements
	private static ElementList elements = new ElementList();
	
	// AttributeList object used for keeping track of attributes
	private static AttributeList attributes = new AttributeList();
	
	// String used for keeping track of the current path
	private String currentPath = "";
	
	// Current element number
	private int currentElement = 0;
	
	// Current location in the xml document
	private int lastIndex = 0;
	
	/**
	 * main - this is the main method that is called from the command line
	 * @params argv[] array of command line arguments
	 */
	public static void main( String argv[] )
	{
		if ( argv.length != 1 )
		{
			System.err.println( "Usage: java s02q2 xmlfilename > htmlfilename" );
			System.exit( 1 );
		}

		String xmlString;
		
		// Use an instance of ourselves as the SAX event handler
		DefaultHandler handler = new s02q2();
		
		// Use the default (non-validating) parser
		SAXParserFactory factory = SAXParserFactory.newInstance();
		
		try
		{
			// Create an input stream from the file specified at the command line
			FileInputStream fin = new FileInputStream( new File( argv[0] ) );
	
			// Remove the white space from the xml file
			xmlString = removeWhiteSpace( fin );
					
			// Create the sax parser
			SAXParser saxParser = factory.newSAXParser();
			
			// Parse the xml string
			saxParser.parse( new ByteArrayInputStream( xmlString.getBytes() ), handler );
			
			// Print the results
			printResults( xmlString );
		} 
		catch( Throwable t )
		{
			t.printStackTrace();
		}
	}
		
	/**
	 * endElement - This method is called each time the parser encounters an
	 *				end element.
	 * @param namespaceURI String representing namespace
	 * @param lname The local name (without prefix), or the empty string if 
					Namespace processing is not being performed.
	 * @param qName The qualified name (with prefix), or the empty string if 
					qualified names are not available.
	 */
	public void endElement( String namespaceURI, String lName, String qName )
		throws SAXException
	{		
		// Retrieve the element from the element list
		XMLElement element = elements.getElement( currentElement );
		
		// Set the end index for the element
		element.setEndIndex( locator.getColumnNumber() );
		
		// Put the element back in the elment list
		elements.setElement( element, currentElement );
		
		// Set the current element to the previous element
		currentElement = element.getPreviousElementID();
		
		// Remove this element from the current path string
		currentPath = currentPath.substring( 0, currentPath.lastIndexOf( '#' ) );
		
		// set the lastIndex to point to the next character
		lastIndex = locator.getColumnNumber() + 1;		
	}	
	
	/**
	 * printAttributeTable - Prints the attribute table
	 */
	public static void printAttributeTable()
	{
		// Print the table header
		System.out.println( "<h3>Attribute Table</h3>" );
		System.out.println( "<table border=\"1\"><tr><td>Name</td>" );
		System.out.println( "<td>Value</td><td>Element Id</td>" );
		System.out.println( "<td>Element Name</td><td>Path Id</td></tr>" );
		
		XMLAttribute attribute;
		
		// Retrieve the attributes as an enumeration
		Enumeration enum = attributes.getElements();
		
		// Loop throught the enumeration
		while( enum.hasMoreElements() )
		{
			// Retrieve the next element
			attribute = (XMLAttribute)enum.nextElement();
			
			// Print out the table row for this element
			System.out.println( "<tr><td>" + 
								attribute.getName() + "</td><td>" + 
								attribute.getValue() + "</td><td>" + 
								attribute.getElementID() + "</td><td>" + 
								attribute.getElementName() + "</td><td>" + 
								attribute.getPathID() + "</td></tr>" );
		}
		
		// Print the end table tag
		System.out.println( "</table>" );
	}
	
	/**
	 * printElementTable - Prints the elements table
	 */
	public static void printElementTable()
	{
		// Print the table header
		System.out.println( "<h3>Element Table</h3>" );
		System.out.print( "<table border=\"1\"><tr><td>Id</td><td>Name</td>" );
		System.out.print( "<td>Path Id</td><td>start index</td>" );
		System.out.println( "<td>end index</td></tr>" );
		
		int elementID = 0;
		XMLElement xmlElement;
		
		// Retrieve the elements as an enumeration
		Enumeration enum = elements.getElements();
		
		// Loop through the enumeration
		while( enum.hasMoreElements() )
		{
			// keep track of the element id
			elementID++;
			
			// Retrieve the next element
			xmlElement = (XMLElement)enum.nextElement();
			
			// Print out the table row for this element
			System.out.println( "<tr><td>" + 
								elementID + "</td><td>" + 
								xmlElement.getName() + "</td><td>" +
								xmlElement.getPathID() + "</td><td>" +
								xmlElement.getStartIndex() + "</td><td>" +
								xmlElement.getEndIndex() + "</td></tr>" );
		}
		
		// Print the end table tag
		System.out.println( "</table>" );
	}
	
	/**
	 * printFooter - prints the html footer information
	 */
	public static void printFooter()
	{
		System.out.println( "</body></html>" );
	}
	
	/**
	 * printHeader - prints the html header information
	 *
	 * @param xmlLength length of the xml source string
	 */
	public static void printHeader( int xmlLength )
	{
		System.out.println( "<html><head><title>XML tables</title></head>" );
		System.out.println( "<body style=\"font-family:verdana;background-color:#ccccff\">" );
		System.out.println( "<h3>Created Time</h3>" + Calendar.getInstance().getTime() );
		System.out.println( "<h3>Source XML</h3>" );
		System.out.println( "xml size => " + xmlLength + "<p>" );
	}
	
	/**
	 * printPathTable - prints the path table
	 */
	public static void printPathTable()
	{
		String path;
		int pathId = 0;
		
		// Print the table header information
		System.out.println( "<h3>Path Table</h3>" );
		System.out.println( "<table border=\"1\"><tr><td>Id</td>" );
		System.out.println( "<td>Path Expressions</td></tr>" );
		
		// Retrieve the paths as an enumeration
		Enumeration enum = paths.getPaths();
		
		// Loop through the enumeration
		while( enum.hasMoreElements() )
		{
			// Keep track of the path id
			pathId++;
			
			// retrieve the next path
			path = (String)enum.nextElement();
			
			// print the table row for this path
			System.out.println( "<tr><td>" + pathId + 
								"</td><td>" + path + "</td></tr>" );
		}
		
		// Print the end table tag
		System.out.println( "</table>" );
	}
	
	/**
	 * This method prints the results of the xml parsing by calling several 
	 * sub print methods.
	 * @param xmlString String representing the xml document with white space 
	 *                  already removed.
	 */
	public static void printResults( String xmlString )
	{
		try
		{
			// Print the html header information
			printHeader( xmlString.length() );
			
			// Print the xml source
			printXMLSource( xmlString );
						
			// Print the path table
			printPathTable();
			
			// Print the element table
			printElementTable();
			
			// Print the attribute table
			printAttributeTable();
			
			// Print the html footer
			printFooter();		
		}
		catch( Throwable t )
		{
			System.out.println( "Error printing results." );
			t.printStackTrace();
		}
	}
	
	/**
	 * printXMLSource - prints the xml source string while replacing any < with
	 *					&lt;
	 * @param xmlSource XML source string
	 */
	public static void printXMLSource( String xmlSource )
	{
		String encodedXML;
		
		// Regular expression pattern
		Pattern lessThan = Pattern.compile( "<" );
		
		// Match the pattern with the xml source
		Matcher matcher = lessThan.matcher( xmlSource );
		
		// Replace the matches with &lt;
		encodedXML = matcher.replaceAll( "&lt;" );
		
		// Print out the results
		System.out.println( "<div style=\"border:1px solid black; padding:5px\">" +
							encodedXML + "</div>" );
	}
	
	/**
	 * removeWhiteSpace - removes white space from an xml string
	 * @param fin File Input Stream
	 * @return String xml string with white space removed
	 */
	public static String removeWhiteSpace( FileInputStream fin )
	{
		// Regular expression patterns
		Pattern leadSpaces = Pattern.compile( "^\\s*<" );
		Pattern endSpaces = Pattern.compile( ">\\s*$" );
		Pattern middleSpaces = Pattern.compile( ">\\s+<" );
		
		// Regular expression matcher
		Matcher matcher;
		
		// String buffer for holder results
		StringBuffer xmlSource = new StringBuffer();
		
		// Input line
		String line;
		
		try
		{
			BufferedReader reader = new BufferedReader( new InputStreamReader( fin ) );
		
			// read 1 line of data
			line = reader.readLine();
		
			// check to see if it is null
			while( line != null )
			{
				// match and remove whitespace from the beginning of a line
				matcher = leadSpaces.matcher( line );
				line = matcher.replaceAll( "<" );
				
				// match and remove whitespace from the end of a line
				matcher = endSpaces.matcher( line );
				line = matcher.replaceAll( ">" );
				
				// match and remove whitespace from between elements
				matcher = middleSpaces.matcher( line );
				line = matcher.replaceAll( "><" );
				
				// append the line to the results string buffer
				xmlSource.append( line );
				
				// read in another line
				line = reader.readLine();
			}
		}
		catch( IOException ioe )
		{
			ioe.printStackTrace();
		}
		
		// return the result string
		return xmlSource.toString();
	}

	/**
	 * setDocumentLocator - Receives a Locator object for document events.
	 * @param locator A locator for all SAX document events.
	 */
	public void setDocumentLocator( Locator locator )
	{
		this.locator = locator;
	}
	
	/**
	 * startElement - this method is called everytime the parser ecounters
					a start element
	 * @param namespaceURI String representing namespace
	 * @param lname The local name (without prefix), or the empty string if 
					Namespace processing is not being performed.
	 * @param qName The qualified name (with prefix), or the empty string if 
					qualified names are not available.
	 * @param attrs The specified or defaulted attributes.
	 */
	public void startElement( String namespaceURI, String lName,
								String qName, Attributes attrs )
		throws SAXException
	{
		int pathID;
		String eName = lName; // element name
		if ( "".equals( eName ) ) 
		{
			eName = qName; // namespaceAware = false
		}
		
		// append to the current path
		currentPath += "#/" + eName;
		
		// add the path to the pathlist
		pathID = paths.addPath( currentPath );
				
		// add the element to the element list
		currentElement = elements.addElement( pathID, eName, 
												lastIndex, 
												currentElement );
		
		// set the lastIndex to point to the current spot
		lastIndex = locator.getColumnNumber();
				
		// If this element contains attributes
		if ( attrs != null ) 
		{
			String aName;
			String aValue;
			
			// Loop through the attributes
			for ( int i = 0; i < attrs.getLength(); i++ ) 
			{
				aName = attrs.getLocalName( i ); // Attr name
				aValue = attrs.getValue( i ); // Attr value
				
				if ( "".equals( aName ) ) 
				{
					aName = attrs.getQName( i );
				}
				
				// Add the path to the path list
				pathID = paths.addPath( currentPath + "#/@" + aName );
				
				attributes.addAttribute( aName, aValue, currentElement, eName, pathID ); 
			}
		}
	}
}